local d = PlayerDamage.init
function PlayerDamage:init(unit)
	d(self, unit)
	
	self._headshot_hot_stack = {}
	self._armour_headshot_hot_stack = {}
	self._doh_headshot_data = tweak_data.upgrades.headshot_hot_data
	self._doh_armor_headshot_data = tweak_data.upgrades.armour_headshot_hot_data
	self._apply_mantra_hot = false
end

function PlayerDamage:add_standing_heal()
	self._apply_mantra_hot = true
end

function PlayerDamage:add_headshot_to_hot(urgency)
	if self:need_revive() or self:dead() or self._check_berserker_done then
		return
	end
	if not urgency then
		table.insert(self._headshot_hot_stack, {
			next_tick = TimerManager:game():time() + (self._doh_headshot_data.tick_time or 1),
			ticks_left = (self._doh_headshot_data.total_ticks or 1)
		})
		table.sort(self._headshot_hot_stack, function(x, y)
			return x.next_tick < y.next_tick
		end)
	else
		local current_armor = self:get_real_armor()
		if current_armor ~= 0 then
			return
		end
		table.insert(self._armour_headshot_hot_stack, {
			next_tick = TimerManager:game():time() + (self._doh_armor_headshot_data.tick_time or 1),
			ticks_left = (self._doh_armor_headshot_data.total_ticks or 1)
		})
		table.sort(self._armour_headshot_hot_stack, function(x, y)
			return x.next_tick < y.next_tick
		end)
	end
end
function PlayerDamage:add_armor_headshot_to_hot()
	if self:need_revive() or self:dead() or self._check_berserker_done then
		return
	end
	local current_armor = self:get_real_armor()
	if current_armor ~= 0 then
		return
	end
	table.insert(self._armour_headshot_hot_stack, {
		next_tick = TimerManager:game():time() + (self._doh_armor_headshot_data.tick_time or 1),
		ticks_left = (self._doh_armor_headshot_data.total_ticks or 1)
	})
	table.sort(self._armour_headshot_hot_stack, function(x, y)
		return x.next_tick < y.next_tick
	end)
end



local odowned = PlayerDamage.on_downed
function PlayerDamage:on_downed()
	odowned(self)
	self._headshot_hot_stack = {}
	self._armour_headshot_hot_stack = {}
end


local update_hp_regen = PlayerDamage._upd_health_regen
function PlayerDamage:_upd_health_regen(t, dt)
	update_hp_regen(self, t, dt)
	if self._apply_mantra_hot then
		self._apply_mantra_hot = false
		self:restore_health(managers.player:upgrade_value("player", "standing_hot", 0), true)
	end
	if 0 < #self._headshot_hot_stack then
		repeat
			local next_doh_head = self._headshot_hot_stack[1]
			local done_head = not next_doh_head or next_doh_head.next_tick > TimerManager:game():time()
			if not done_head then
				local regen_rate_head = managers.player:upgrade_value("player", "headshot_hot", 0)
				self:restore_health(regen_rate_head, true)
				next_doh_head.ticks_left = next_doh_head.ticks_left - 1
				if next_doh_head.ticks_left == 0 then
					table.remove(self._headshot_hot_stack, 1)
				else
					next_doh_head.next_tick = next_doh_head.next_tick + (self._doh_headshot_data.tick_time or 1)
				end
				table.sort(self._headshot_hot_stack, function(x, y)
					return x.next_tick < y.next_tick
				end)
			end
		until done_head
	end
	if 0 < #self._armour_headshot_hot_stack then
		repeat
			local next_doh_armor = self._armour_headshot_hot_stack[1]
			local done_armor = not next_doh_armor or next_doh_armor.next_tick > TimerManager:game():time()
			if not done_armor then
				local regen_rate_armor = managers.player:upgrade_value("player", "armour_headshot_hot", 0)
				self:restore_health(regen_rate_armor, true)
				next_doh_armor.ticks_left = next_doh_armor.ticks_left - 1
				if next_doh_armor.ticks_left == 0 then
					table.remove(self._armour_headshot_hot_stack, 1)
				else
					next_doh_armor.next_tick = next_doh_armor.next_tick + (self._doh_armor_headshot_data.tick_time or 1)
				end
				table.sort(self._armour_headshot_hot_stack, function(x, y)
					return x.next_tick < y.next_tick
				end)
			end
		until done_armor
	end
end

























function PlayerDamage:_check_bleed_out(can_activate_berserker, ignore_movement_state)
	if self:get_real_health() == 0 and not self._check_berserker_done then
		if self._unit:movement():zipline_unit() then
			self._bleed_out_blocked_by_zipline = true
			return
		end
		if not ignore_movement_state and self._unit:movement():current_state():bleed_out_blocked() then
			self._bleed_out_blocked_by_movement_state = true
			return
		end
		local time = Application:time()
		if not self._block_medkit_auto_revive and time > self._uppers_elapsed + self._UPPERS_COOLDOWN then
			local auto_recovery_kit = FirstAidKitBase.GetFirstAidKit(self._unit:position())
			if auto_recovery_kit then
				auto_recovery_kit:take(self._unit)
				self._unit:sound():play("pickup_fak_skill")
				self._uppers_elapsed = time
				return
			end
		end
		if can_activate_berserker and not self._check_berserker_done then
			local has_berserker_skill = managers.player:has_category_upgrade("temporary", "berserker_damage_multiplier")
			if has_berserker_skill and not self._disable_next_swansong then
				managers.hud:set_teammate_condition(HUDManager.PLAYER_PANEL, "mugshot_swansong", managers.localization:text("debug_mugshot_downed"))
				managers.player:activate_temporary_upgrade("temporary", "berserker_damage_multiplier")
				self._current_state = nil
				self._check_berserker_done = true
				if alive(self._interaction:active_unit()) and not self._interaction:active_unit():interaction():can_interact(self._unit) then
					self._unit:movement():interupt_interact()
				end
				self._listener_holder:call("on_enter_swansong")
			end
			self._disable_next_swansong = nil
		end
		self._hurt_value = 0.2
		self._damage_to_hot_stack = {}
		
		
		self._headshot_hot_stack = {}
		self._armour_headshot_hot_stack = {}
		
		
		managers.environment_controller:set_downed_value(0)
		SoundDevice:set_rtpc("downed_state_progression", 0)
		if not self._check_berserker_done or not can_activate_berserker then
			self._revives = Application:digest_value(Application:digest_value(self._revives, false) - 1, true)
			self._check_berserker_done = nil
			managers.environment_controller:set_last_life(Application:digest_value(self._revives, false) <= 1)
			if Application:digest_value(self._revives, false) == 0 then
				self._down_time = 0
			end
			self._bleed_out = true
			self._current_state = nil
			managers.player:set_player_state("bleed_out")
			self._critical_state_heart_loop_instance = self._unit:sound():play("critical_state_heart_loop")
			self._slomo_sound_instance = self._unit:sound():play("downed_slomo_fx")
			self._bleed_out_health = Application:digest_value(tweak_data.player.damage.BLEED_OUT_HEALTH_INIT * managers.player:upgrade_value("player", "bleed_out_health_multiplier", 1), true)
			self:_drop_blood_sample()
			self:on_downed()
		end
	elseif not self._said_hurt and 0.2 > self:get_real_health() / self:_max_health() then
		self._said_hurt = true
		PlayerStandard.say_line(self, "g80x_plu")
	end
end